/*
 * Copyright (C) 2001-2004 The Honeynet Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by The Honeynet Project.
 * 4. The name "The Honeynet Project" may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * This code was originally written by Blake R. Watts  (blake@blakewatts.com)
 * and graciously donated to the The Honeynet Project.
 */

#include <ntddk.h>
#include "consolespy.h"
#include "sebek.h"
#include "logging.h"
#include "util.h"
#include "tib.h"

static PZWREQUESTWAITREPLYPORT s_fnZwRequestWaitReplyPort;
static PZWCLOSE s_fnZwClose;
static PZWSECURECONNECTPORT s_fnZwSecureConnectPort;
static PZWWRITEFILE s_fnZwWriteFile;
static PZWREADFILE s_fnZwReadFile;

static PONCONSOLEIO s_fnOnConsoleWrite;
static PONCONSOLEIO s_fnOnConsoleRead;

static LIST_ENTRY s_PortHandleList;
static KSPIN_LOCK s_PortHandleListLock;
static BOOLEAN s_PortDataInitialized;

NTSTATUS
UninitConsoleSpy()
{
   if (s_fnZwRequestWaitReplyPort)
   {
      if(!SetSystemCallIndex(SYSCALL_INDEX_ZWREQUESTWAITREPLYPORT, s_fnZwRequestWaitReplyPort, 0))
				return STATUS_UNSUCCESSFUL;

      if(!SetSystemCallIndex(SYSCALL_INDEX_ZWCLOSE, (PVOID)s_fnZwClose, 0))
				return STATUS_UNSUCCESSFUL;

      if(!SetSystemCallIndex(SYSCALL_INDEX_ZWSECURECONNECTPORT, (PVOID)s_fnZwSecureConnectPort, 0))
				return STATUS_UNSUCCESSFUL;

			if(!SetSystemCallIndex(SYSTEMSERVICE(ZwReadFile), (PVOID)s_fnZwReadFile, NULL))
				return STATUS_UNSUCCESSFUL;

			if(!SetSystemCallIndex(SYSTEMSERVICE(ZwWriteFile), (PVOID)s_fnZwWriteFile, NULL))
				return STATUS_UNSUCCESSFUL;
   }

	 return STATUS_SUCCESS;
}

void
OnConsoleWrite(ULONG ProcessId, const PANSI_STRING str)
{
	if(!str)
		return;

	DBGOUT(("OnConsoleWrite(): %d bytes read -> '%s'\n", str->Length, str->Buffer));
	LogData(ProcessId, str->Buffer, str->Length);
}

void
OnConsoleRead(ULONG ProcessId, const PANSI_STRING str)
{
	if(!str)
		return;

	DBGOUT(("OnConsoleRead(): %d bytes read -> '%s'\n", str->Length, str->Buffer));
	LogData(ProcessId, str->Buffer, str->Length);
}


// ----------------------
// Special Event Handlers
//
static void
OnCsrWriteDataPre(ULONG ProcessId, PCSR_CONSOLE_WRITE_MESSAGE Message, ULONG VirtualOffset)
{
   PVOID WriteStringPtr = 0;
   USHORT cbSize = 0;
   ANSI_STRING as;

	 if(!Message)
		 return;

   //
   // Is the buffer inline with the Message?
   //
   __try
   {
      if (Message->WriteInfo.MessageBufferPtr != Message->WriteInfo.MessageBuffer)
      {
         PUCHAR Offset;

         //
         // Nope - Let's translate the address into something usable and verifiable
         //
         Offset = (PUCHAR)((ULONG)Message->WriteInfo.MessageBufferPtr - VirtualOffset);
         if (MmIsAddressValid((PVOID)Offset) && 
            MmIsAddressValid((PVOID)(&Offset[Message->WriteInfo.MessageBufferSize])))
         {
            WriteStringPtr = (PVOID)Offset;
            cbSize = (USHORT)Message->WriteInfo.MessageBufferSize;
         }
      }
      else
      {
         WriteStringPtr = (PVOID)Message->WriteInfo.MessageBuffer;
         cbSize = (USHORT)(min(Message->WriteInfo.MessageBufferSize, CONSOLE_WRITE_INFO_MESSAGE_BUFFER_SIZE));
      }

      if (!WriteStringPtr) // bad data
         return;

      if (Message->WriteInfo.Unicode)
      {
         UNICODE_STRING us;

         us.Buffer = (wchar_t*)WriteStringPtr;
         us.Length = cbSize;
         us.MaximumLength = cbSize;

         RtlUnicodeStringToAnsiString(&as, &us, TRUE);
      }
      else
      {
         as.Length = (USHORT)Message->WriteInfo.MessageBufferSize;
         as.MaximumLength = (USHORT)Message->WriteInfo.MessageBufferSize;
         as.Buffer = (char*)WriteStringPtr;
      }

      if (s_fnOnConsoleWrite)
         s_fnOnConsoleWrite(ProcessId, &as);

      // free allocated ansi buffer if needed
      if (Message->WriteInfo.Unicode)
         RtlFreeAnsiString(&as);

   } __except(EXCEPTION_EXECUTE_HANDLER)
   {
   }
}

static void
OnCsrReadDataPost(ULONG ProcessId, PCSR_CONSOLE_READ_MESSAGE Message, ULONG VirtualOffset)
{
   PVOID ReadStringPtr = 0;
   ANSI_STRING as;

	 if(!Message)
		 return;

   //
   // Is the buffer inline with the Message?
   //
   __try
   {
      if (Message->ReadInfo.MessageBufferPtr != Message->ReadInfo.MessageBuffer)
      {
         PUCHAR Offset;

         //
         // Nope - Let's translate the address into something usable and verifiable
         //
         Offset = (PUCHAR)((ULONG)Message->ReadInfo.MessageBufferPtr - VirtualOffset);
         if (MmIsAddressValid((PVOID)Offset) && 
            MmIsAddressValid((PVOID)(&Offset[Message->ReadInfo.MessageBufferSize])))
         {
            ReadStringPtr = (PVOID)Offset;
         }
      }
      else
      {
         ReadStringPtr = (PVOID)Message->ReadInfo.MessageBuffer;
      }

      if (!ReadStringPtr) // bad data
         return;

      if (Message->ReadInfo.Unicode)
      {
         UNICODE_STRING us;

         us.Buffer = (wchar_t*)ReadStringPtr;
         us.Length = (USHORT)Message->ReadInfo.NumberOfCharsToRead;
         us.MaximumLength = (USHORT)Message->ReadInfo.NumberOfCharsToRead;

         RtlUnicodeStringToAnsiString(&as, &us, TRUE);
      }
      else
      {
         as.Length = (USHORT)Message->ReadInfo.NumberOfCharsToRead;
         as.MaximumLength = (USHORT)Message->ReadInfo.NumberOfCharsToRead;
         as.Buffer = (char*)ReadStringPtr;
      }

      if (s_fnOnConsoleRead)
         s_fnOnConsoleRead(ProcessId, &as);

      // free allocated ansi buffer if needed
      if (Message->ReadInfo.Unicode)
         RtlFreeAnsiString(&as);

   } __except(EXCEPTION_EXECUTE_HANDLER)
   {
   }
}

static
void
//lint -e715
OnCsrWriteDataPost(ULONG ProcessId, PCSR_CONSOLE_WRITE_MESSAGE Message, ULONG VirtualOffset)
{
	return;
} // Nothing to do

static
void
//lint -e715
OnCsrReadDataPre(ULONG ProcessId, PCSR_CONSOLE_READ_MESSAGE Message, ULONG VirtualOffset)
{
	return;
} // Nothing to do


// ----------------
// Hooked Syscalls
//
static
NTSTATUS 
NTAPI 
OnZwRequestWaitReplyPort(
   IN HANDLE PortHandle,
   IN PPORT_MESSAGE RequestMessage,
   OUT PPORT_MESSAGE ReplyMessage
   )
{
   NTSTATUS status;
   BOOLEAN bDiscard;
	 
	 bDiscard = TRUE; 
   
   InterlockedIncrement((PLONG)&s_ulCallCount);

   if(RequestMessage != NULL && IsCsrssPortHandle((ULONG)PsGetCurrentProcessId(), PortHandle))
   {
      PCSRSS_MESSAGE pCsrMsg = (PCSRSS_MESSAGE)((PPORT_MESSAGE)RequestMessage + 1);
      
      ULONG VirtualOffset = GetVirtualOffsetFromHandle((ULONG)PsGetCurrentProcessId(), PortHandle);

      if (pCsrMsg->OpCode == OPCODE_WRITE_CONSOLE)
      {
         PCSR_CONSOLE_WRITE_MESSAGE pWriteMsg = (PCSR_CONSOLE_WRITE_MESSAGE)RequestMessage;
         
         OnCsrWriteDataPre((ULONG)PsGetCurrentProcessId(), pWriteMsg, VirtualOffset);
         status = s_fnZwRequestWaitReplyPort(PortHandle, RequestMessage, ReplyMessage);
         OnCsrWriteDataPost((ULONG)PsGetCurrentProcessId(), pWriteMsg, VirtualOffset);
         
         bDiscard = FALSE;
      }
      else if (pCsrMsg->OpCode == OPCODE_READ_CONSOLE)
      {
         PCSR_CONSOLE_READ_MESSAGE pReadMsg = (PCSR_CONSOLE_READ_MESSAGE)RequestMessage;
         
         OnCsrReadDataPre((ULONG)PsGetCurrentProcessId(), pReadMsg, VirtualOffset);
         status = s_fnZwRequestWaitReplyPort(PortHandle, RequestMessage, ReplyMessage);
         OnCsrReadDataPost((ULONG)PsGetCurrentProcessId(), pReadMsg, VirtualOffset);

         bDiscard = FALSE;
      }
   }

   if (bDiscard)
      status = s_fnZwRequestWaitReplyPort(PortHandle, RequestMessage, ReplyMessage);
   
   InterlockedDecrement((PLONG)&s_ulCallCount);

   return status;
}


static
NTSTATUS
NTAPI
OnZwSecureConnectPort(
   OUT PHANDLE PortHandle,
   IN PUNICODE_STRING PortName,
   IN PSECURITY_QUALITY_OF_SERVICE SecurityQos,
   IN OUT PPORT_SECTION_WRITE WriteSection OPTIONAL,
   IN PSID ServerSid OPTIONAL,
   IN OUT PPORT_SECTION_READ ReadSection OPTIONAL,
   OUT PULONG MaxMessageSize OPTIONAL,
   IN OUT PVOID ConnectData OPTIONAL,
   IN OUT PULONG ConnectDataLength OPTIONAL
   )
{
   NTSTATUS status;
   
   InterlockedIncrement((PLONG)&s_ulCallCount);

   status = s_fnZwSecureConnectPort(PortHandle, PortName, SecurityQos, WriteSection, 
      ServerSid, ReadSection, MaxMessageSize, ConnectData, ConnectDataLength);

   if (NT_SUCCESS(status))
   {
      if(PortName != NULL && MmIsAddressValid(PortName) 
				&& PortName->Buffer != NULL && MmIsAddressValid(PortName->Buffer) 
				&& WriteSection != NULL && MmIsAddressValid(WriteSection))
      {
         if (wcsncmp(PortName->Buffer, L"\\Windows\\ApiPort", PortName->Length / sizeof(WCHAR)) == 0)
         {
            ULONG VirtualOffset = (ULONG)WriteSection->TargetViewBase - (ULONG)WriteSection->ViewBase;
						//lint -e613
            InsertCsrssPortHandle((ULONG)PsGetCurrentProcessId(), (HANDLE)(*((PULONG)PortHandle)), VirtualOffset);
         }
      }
   }


   InterlockedDecrement((PLONG)&s_ulCallCount);

   return status;
}


static
NTSTATUS
NTAPI
OnZwClose(
   IN HANDLE Handle
   )
{
   NTSTATUS status;
   
   InterlockedIncrement((PLONG)&s_ulCallCount);

   if (IsCsrssPortHandle((ULONG)PsGetCurrentProcessId(), Handle))
      RemoveCsrssPortHandle((ULONG)PsGetCurrentProcessId(), Handle);

   status = s_fnZwClose(Handle);

   InterlockedDecrement((PLONG)&s_ulCallCount);

   return status;
}

/*
 *	Ok, a little documentation on how LogIfStdHandle works.
 *
 *	First, we retreive the address of the TIB for the thread calling this function.
 *	This should be from a hooked service sucha s ZwReadFile and ZwWriteFile.
 *	Then we get the address of the PEB, and then the Environment block inside the PEB.
 *	The Environment block contains HANDLEs to STDIN, STDOUT, and STDERR.
 *
 *	We check to see if the WriteFile is attempting to write to one of those handles.
 *	If WriteFile is attempting to write to one of those handles then the process was passed
 *	a HANDLE instead of using the standard console subsystem pseudo handles.
 *
 *	This should catch any exploits which pass a socket or file directly to CreateProcess().
 */
static VOID
LogIfStdHandle(
		IN CONST HANDLE FileHandle,
		IN PVOID Buffer,
		IN ULONG Length
		)
{
	PTIB pTIB = NULL;

	/*
	 *	Check that the call is from UserMode and double check we are not being called
	 *	at too high of an IRQL.
	 */
	if(ExGetPreviousMode() == UserMode && KeGetCurrentIrql() < DISPATCH_LEVEL) {
		__asm
    {
        mov     EAX, FS:[18h]		; Store TIB in 32bit register so we don't have to mess with segments.
				mov     [pTIB], EAX			; Push the address into our pointer
    }

		//lint -e774
		if(pTIB) {
			if(pTIB->pPEB && pTIB->pPEB->pEnv) {
				if(FileHandle == pTIB->pPEB->pEnv->hStdOut || FileHandle == pTIB->pPEB->pEnv->hStdIn || FileHandle == pTIB->pPEB->pEnv->hStdErr)
					LogData(pTIB->processID, (PBYTE)Buffer, Length);
			}
		}
	}
}


/*
 *	Read above documentation for LogIfStdHandle on how this function works.
 */
static NTSTATUS
NTAPI
OnZwReadFile(
  	IN HANDLE FileHandle,
		IN HANDLE Event OPTIONAL,
		IN PIO_APC_ROUTINE ApcRoutine OPTIONAL,
		IN PVOID ApcContext OPTIONAL,
		OUT PIO_STATUS_BLOCK IoStatusBlock,
		IN PVOID Buffer,
		IN ULONG Length,
		IN PLARGE_INTEGER ByteOffset OPTIONAL,
		IN PULONG Key OPTIONAL
		)
{
	NTSTATUS status;
	
	InterlockedIncrement((PLONG)&s_ulCallCount);

	status = s_fnZwReadFile(FileHandle, Event, ApcRoutine, ApcContext, IoStatusBlock, Buffer, Length, ByteOffset, Key);

	if(status == STATUS_SUCCESS) 
		LogIfStdHandle(FileHandle, Buffer, Length);

	InterlockedDecrement((PLONG)&s_ulCallCount);

	return status;
}

/*
 *	Read above documentation for LogIfStdHandle on how this function works.
 */
static NTSTATUS
NTAPI
OnZwWriteFile(
  	IN HANDLE FileHandle,
		IN HANDLE Event OPTIONAL,
		IN PIO_APC_ROUTINE ApcRoutine OPTIONAL,
		IN PVOID ApcContext OPTIONAL,
		OUT PIO_STATUS_BLOCK IoStatusBlock,
		IN PVOID Buffer,
		IN ULONG Length,
		IN PLARGE_INTEGER ByteOffset OPTIONAL,
		IN PULONG Key OPTIONAL
		)
{
	NTSTATUS status;
	
	InterlockedIncrement((PLONG)&s_ulCallCount);

	status = s_fnZwWriteFile(FileHandle, Event, ApcRoutine, ApcContext, IoStatusBlock, Buffer, Length, ByteOffset, Key);

	if(status == STATUS_SUCCESS) 
		LogIfStdHandle(FileHandle, Buffer, Length);

	InterlockedDecrement((PLONG)&s_ulCallCount);

	return status;
}

// -------------------------------------
// Port Handle Data Structure Management
//
NTSTATUS
InitConsoleSpy(PONCONSOLEIO fnWrite, PONCONSOLEIO fnRead)
{
	s_fnOnConsoleWrite = fnWrite;
	s_fnOnConsoleRead = fnRead;

	InitializeListHead(&s_PortHandleList);
	KeInitializeSpinLock(&s_PortHandleListLock);

	s_PortDataInitialized = TRUE;

	// Redirect syscalls
	if(!SetSystemCallIndex(SYSCALL_INDEX_ZWREQUESTWAITREPLYPORT, (PVOID)OnZwRequestWaitReplyPort, (PVOID*)&s_fnZwRequestWaitReplyPort))
		return STATUS_UNSUCCESSFUL;

	if(!SetSystemCallIndex(SYSCALL_INDEX_ZWCLOSE, (PVOID)OnZwClose, (PVOID*)&s_fnZwClose))
		return STATUS_UNSUCCESSFUL;

	if(!SetSystemCallIndex(SYSCALL_INDEX_ZWSECURECONNECTPORT, (PVOID)OnZwSecureConnectPort, (PVOID*)&s_fnZwSecureConnectPort))
		return STATUS_UNSUCCESSFUL;
	
	if(!SetSystemCallIndex(SYSTEMSERVICE(ZwReadFile), OnZwReadFile, (PVOID *)&s_fnZwReadFile))
		return STATUS_UNSUCCESSFUL;

	if(!SetSystemCallIndex(SYSTEMSERVICE(ZwWriteFile), OnZwWriteFile, (PVOID *)&s_fnZwWriteFile))
		return STATUS_UNSUCCESSFUL;

	DBGOUT(("InitConsoleSpy\n"));
	return STATUS_SUCCESS;
}

static 
PCSRSS_PORT_HANDLE_ENTRY
GetCsrssHandleEntry(
   IN ULONG ProcessId,
   IN CONST HANDLE PortHandle
   )
{
   if (!IsListEmpty(&s_PortHandleList))
   {
      PCSRSS_PORT_HANDLE_ENTRY pCurEntry = 0;

      pCurEntry = (PCSRSS_PORT_HANDLE_ENTRY)s_PortHandleList.Flink;
      do
      {
         if (pCurEntry->ProcessId == ProcessId && pCurEntry->PortHandle == PortHandle)
            return pCurEntry;

         pCurEntry = pCurEntry = (PCSRSS_PORT_HANDLE_ENTRY)pCurEntry->ListEntry.Flink;

      } while (pCurEntry != (PCSRSS_PORT_HANDLE_ENTRY)&s_PortHandleList);
   }

   return 0;
}

static ULONG
GetVirtualOffsetFromHandle(
   IN ULONG ProcessId,
   IN HANDLE PortHandle
   )
{
   KIRQL Irql;
   ULONG VirtualOffset = 0;
   PCSRSS_PORT_HANDLE_ENTRY Entry;

   if (!s_PortDataInitialized)
      return FALSE;

   KeAcquireSpinLock(&s_PortHandleListLock, &Irql);
   
   Entry = GetCsrssHandleEntry(ProcessId, PortHandle);
   if (Entry)
      VirtualOffset = Entry->VirtualOffset;

   KeReleaseSpinLock(&s_PortHandleListLock, Irql);

   return VirtualOffset;
}


static BOOLEAN 
IsCsrssPortHandle(
   IN ULONG ProcessId,
   IN HANDLE PortHandle
   )
{
   KIRQL Irql;
   BOOLEAN bCsrssHandle = FALSE;
   PCSRSS_PORT_HANDLE_ENTRY Entry;

   if (!s_PortDataInitialized)
      return FALSE;

   KeAcquireSpinLock(&s_PortHandleListLock, &Irql);
   
   Entry = GetCsrssHandleEntry(ProcessId, PortHandle);
   if (Entry)
      bCsrssHandle = TRUE;

   KeReleaseSpinLock(&s_PortHandleListLock, Irql);

   return bCsrssHandle;
}


static VOID
InsertCsrssPortHandle(
   IN ULONG ProcessId,
   IN HANDLE PortHandle,
   IN ULONG VirtualOffset
   )
{
   PCSRSS_PORT_HANDLE_ENTRY Entry;
   KIRQL Irql;

   if (!s_PortDataInitialized)
      return;

   Entry = (PCSRSS_PORT_HANDLE_ENTRY)ExAllocatePool(PagedPool, 
      sizeof(CSRSS_PORT_HANDLE_ENTRY));

   if (!Entry)
      return;

   DBGOUT(("InsertCsrssPortHandle(PID:%d; PortHandle:%xh; Offset:0x%08x\n", 
      ProcessId, PortHandle, VirtualOffset));

   Entry->PortHandle = PortHandle;
   Entry->ProcessId = ProcessId;
   Entry->VirtualOffset = VirtualOffset;

   KeAcquireSpinLock(&s_PortHandleListLock, &Irql);
   InsertHeadList(&s_PortHandleList, &Entry->ListEntry);
   KeReleaseSpinLock(&s_PortHandleListLock, Irql);
}


static VOID
RemoveCsrssPortHandle(
   IN ULONG ProcessId,
   IN HANDLE PortHandle
   )
{
   PCSRSS_PORT_HANDLE_ENTRY Entry;
   KIRQL Irql;

   if (!s_PortDataInitialized)
      return;

   KeAcquireSpinLock(&s_PortHandleListLock, &Irql);

   Entry = GetCsrssHandleEntry(ProcessId, PortHandle);

   if (Entry)
   {
      DBGOUT(("RemoveCsrssPortHandle(PID:%d; PortHandle:%xh\n", ProcessId, PortHandle));

      RemoveEntryList(&Entry->ListEntry);
      ExFreePool(Entry);
   }
   
   KeReleaseSpinLock(&s_PortHandleListLock, Irql);
}
